/*
 * vifd/libvif.c
 *
 * Copyright (C) 2008-2009 Reto Gantenbein
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include "libvif.h"
#include "vifd_print.h"

int
msgq_initialize(int key)
{
    int id;
    int msgq_flags = IPC_CREAT | 0664; // create read/write queue

    if ((id = msgget(key, msgq_flags)) < 0)
    {
        perror("msgget");
        exit(1);
    } else {
        return id;
    }
}

int
msgq_receive(int id, void * buffer, int size, int type)
{
    int rtrn;

    if ((rtrn = msgrcv(id, buffer, size, type, 0)) < 0)
    {
        perror("msgrcv"); // raise error, but continue execution
    }
    return rtrn;
}

int
msgq_delete(int id)
{
    return msgctl(id, IPC_RMID, NULL);
}

int
shmem_initialize(int key, int bytes)
{
    int id;
    int shmem_flags = IPC_CREAT | IPC_EXCL | 0600;

    if ((id = shmget(key, bytes, shmem_flags)) == -1)
    {
        if (errno != EEXIST)
        {
            perror("shmget");
        } else {  // if already exists, reattach and clean it
            if ((id = shmget(key, 0, 0600)) > -1)
            {
                memset(shmat(id, NULL, 0), 0, bytes);
            }
        }
    }

    return id;
}

char *
shmem_attach(int id, char * addr)
{
    char * return_addr;
    if ((return_addr = shmat(id, addr, 0)) == (char *) -1)
    {
        perror("shmat");
    }
    // evtl. add locking?

    return return_addr;
}

int
shmem_detach(char * addr)
{
    if (shmdt(addr) < 0)
    {
        perror("shmdt");
        return -1;
    }

    return 0;
}

int
shmem_delete(int id)
{
    return shmctl(id, IPC_RMID, NULL);
}

int
set_iw_param_state(struct iw_param * par)
{
    return !(par->disabled == 1);
}

void
create_vif_config(const char *vif_name, int index)
{
    int vif_count, i;

    // initialize wireless configuration in shared memory
    vifd_debug("create_vif_config(): vif_index = %i, vif_name = %s\n", \
        index, vif_name);
    if (index == 0) {
        vif_list[0] = (struct vif_info *) shmat(pm_shm_id, NULL, 0);
        if ((void *) vif_list[0] == (void *) -1)
        {
            perror("vifd_shmat");
            exit(1);
        }
        vifd_debug("create_vif_config(): [%i] %p\n", index, vif_list[0]);
        for (vif_count = 1; vif_count < MAX_NUM_VIF; vif_count++)
        {
            vif_list[vif_count] = vif_list[0] + vif_count;
            vifd_debug("create_vif_config(): [%i] %p\n", \
                vif_count, vif_list[vif_count]);
        }
    }

    // set device name
    strncpy((char *) vif_list[index]->vif_name, vif_name, sizeof(if_name));

    // create wireless config
    struct wireless_info * vif_config = &(vif_list[index]->vif_config);
    vifd_debug("create_vif_config(): device %i : %s, wireless_info = %p\n", \
        index, vif_list[index]->vif_name, vif_config);

    // set wireless name
    strcpy(vif_config->b.name, "IEEE 802.11b");

    // define driver parameter range
    vif_config->range.min_nwid = 0;
    vif_config->range.max_nwid = 0;
    vif_config->range.sensitivity = 3;  // ?
    vif_config->range.we_version_compiled = iw_get_kernel_we_version();
    vif_config->range.we_version_source = 20;
    vif_config->range.num_channels = VIFD_NUM_CHANNEL;
    vif_config->range.num_frequency = VIFD_NUM_FREQ;
    for (i = 0; i < VIFD_NUM_FREQ; i++)
    {
        iw_float2freq(channel_frequency[i] * MEGA, &(vif_config->range.freq[i]));
        vif_config->range.freq[i].i = i;
    }
    vif_config->range.min_retry = 0;
    vif_config->range.max_retry = 255;
    vif_config->range.min_rts = 0;
    vif_config->range.max_rts = 2347;

    // set network id (default: off)
    vif_config->b.nwid.value = 0;
    vif_config->b.nwid.disabled = 1;
    vif_config->b.nwid.fixed = 1;
    vif_config->b.has_nwid = set_iw_param_state(&(vif_config->b.nwid));

    // set AP address (default: none)
    vif_config->has_ap_addr = 0;
    memset((char *) vif_config->ap_addr.sa_data, 0x00, ETH_ALEN);

    // power management settings (default: off)
    vif_config->power.value = 0;
    vif_config->power.disabled = 1;
    vif_config->power.fixed = 1;
    vif_config->has_power = set_iw_param_state(&(vif_config->power));

    // set nickname (default: none)
    vif_config->has_nickname = 0;
    vif_config->nickname[0] = '\0';

    // set default parameters
    set_bitrate(NULL, vif_config);
    set_encoding(NULL, vif_config);
    set_essid(NULL, vif_config);
    set_freq(NULL, vif_config);
    set_frag(NULL, vif_config);
    set_mode(NULL, vif_config);
    set_rts(NULL, vif_config);
    set_sens(NULL, vif_config);
    set_txpower(NULL, vif_config);

    num_vif++;
}

int
get_vif_index(if_name *vif)
{
    int i;

    for (i = 0; i < num_vif; i++) {
        if (strcmp((char *) vif, (char *) &(vif_list[i]->vif_name)) == 0)
        {
            return i;
        }
    }
    vifd_debug("get_vif_index(): device %s not found!\n", (char *) vif);
    return -1;
}

void
handle_iwreq(struct iwreq_msg *message_buf)
{
    int i, rtrn = -1;
    int vif_index = get_vif_index((if_name *) message_buf->iw_request.ifr_name);

    if (vif_index < 0)
    {
        vifd_debug("handle_iwreq(): IPC_FAIL\n");
        message_buf->iwreq_ack = IPC_FAIL;
    } else {
        int max_ioctl = sizeof(ioctl_handler_list) / sizeof(struct ioctl_handler);
        struct wireless_info * vif_config = &(vif_list[vif_index]->vif_config);

        // execute iw_handler from ioctl list
        for (i = 0; i < max_ioctl; i++)
        {
            if (ioctl_handler_list[i].ioctl_num == message_buf->iwreq_property)
            {
                rtrn = (*ioctl_handler_list[i].iw_handler) \
                    (&(message_buf->iw_request), vif_config);

                // set failure flag if unsuccessful
                if (rtrn < 0)
                {
                    vifd_debug("handle_iwreq(): IPC_FAIL\n");
                    message_buf->iwreq_ack = IPC_FAIL;
                } else {
                    message_buf->iwreq_ack = IPC_OK;

                    // inform packetmodeller about update
                    if ((message_buf->iwreq_property & (1 << 0)) == 0)
                    {
                        struct pm_msg pm_notify;
                        pm_notify.mtype = PM_UPDATE;
                        pm_notify.vif_index = vif_index;
                        pm_notify.property = message_buf->iwreq_property;
                        if (msgsnd(pm_msgq_id, &pm_notify, \
                            sizeof(struct pm_msg), 0) < 0)
                            perror("pm_msgsnd");
                    }
                }

                // set data flag
                message_buf->iwreq_ack = message_buf->iwreq_ack \
                    | ioctl_handler_list[i].data_flag;
                break;
            }
        }
    }

    // send ioctl acknowlege
    message_buf->mtype = IW_ACK;
    if (msgsnd(iw_msgq_id, message_buf, IW_MSG_SIZE, 0) < 0)
        perror("iw_msgsnd"); // raise error, but continue execution
}

int
get_name(struct iwreq * request, struct wireless_info * vif_config)
{
    memmove(request->u.name, vif_config->b.name, IFNAMSIZ);
    vifd_debug("SIGCGIWNAME : %s\n", request->u.name);

    return 0;
}

int
get_nwid(struct iwreq * request, struct wireless_info * vif_config)
{
    if (vif_config->b.has_nwid == 0)
    {
        vifd_debug("SIOCGIWNWID : off\n");
        return IWERR_GET_EXT;
    }
    memmove(&(request->u.nwid), &vif_config->b.nwid, sizeof(struct iw_param));
    vifd_debug("SIOCGIWNWID : %c\n", request->u.nwid.value);

    return 0;
}

/** called for SIOCSIWFREQ (argument can be channel or frequency) */
int
set_freq(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)   // set initial frequency (default: channel 0)
    {
        vif_config->b.has_freq = 0;
        vif_config->b.freq = channel_frequency[0] * MEGA;
        vif_config->b.freq_flags = IW_FREQ_AUTO;
    } else {
        vif_config->b.freq_flags = request->u.freq.flags;
        if (vif_config->b.freq_flags == IW_FREQ_AUTO)   // reset frequency
        {
            vif_config->b.has_freq = 0;
            vif_config->b.freq = channel_frequency[0] * MEGA;
        } else {                                // set new frequency
            vif_config->b.has_freq = 1;
            float freq = iw_freq2float(&(request->u.freq));

            // if freq < 1000 we assume it's a channel
            if ((freq < MEGA) && (freq < vif_config->range.num_channels))
            {
                int channel = (int) freq;
                vif_config->b.freq = channel_frequency[channel] * MEGA;
            } else {
                vif_config->b.freq = freq;
            }
        }
    }
    if (vif_config->b.freq > 1000)
    {
        vifd_debug("SIOCSIWFREQ : %.3e GHz\n", vif_config->b.freq / GIGA);
    } else {
        vifd_debug("SIOCSIWFREQ : Channel %.0f\n", vif_config->b.freq);
    }

    return 0;
}

int
get_freq(struct iwreq * request, struct wireless_info * vif_config)
{
    iw_float2freq(vif_config->b.freq, &(request->u.freq));
    if (vif_config->b.freq > 1000)
    {
        vifd_debug("SIOCGIWFREQ : %.3e GHz\n", vif_config->b.freq / GIGA);
    } else {
        vifd_debug("SIOCGIWFREQ : Channel %.0f\n", vif_config->b.freq);
    }

    return 0;
}

int
set_mode(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)                // set initial mode (default: infra)
    {
        vif_config->b.has_mode = 1;
        vif_config->b.mode = IW_MODE_INFRA;
    } else {                                                  // set new mode
        vif_config->b.has_mode = 1;
        vif_config->b.mode = request->u.mode;
    }
    vifd_debug("SIOCSIWMODE : %i\n", vif_config->b.mode);

    return 0;
}

int
get_mode(struct iwreq * request, struct wireless_info * vif_config)
{
    if (vif_config->b.has_mode == 1)
    {
        request->u.mode = vif_config->b.mode;
    } else {
        request->u.mode = IW_MODE_AUTO;
    }
    vifd_debug("SIOCGIWMODE : %i\n", request->u.mode);

    return 0;
}

int
set_sens(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)
    {
        vif_config->sens.value = 0;  //TODO: find out what this means
        vif_config->sens.disabled = 0;
        vif_config->sens.fixed = 1;
    } else {
        memmove(&(vif_config->sens), &(request->u.sens), sizeof(struct iw_param));
    }
    vif_config->has_sens = set_iw_param_state(&(vif_config->sens));
    if (vif_config->sens.value < 0)
        vifd_debug("SIOCSIWSENS : %i dB\n", vif_config->sens.value);
    else
        vifd_debug("SIOCSIWSENS : %i\n", vif_config->sens.value);

    return 0;
}

int
get_sens(struct iwreq * request, struct wireless_info * vif_config)
{
    memmove(&(request->u.sens), &(vif_config->sens), sizeof(struct iw_param));
    if (vif_config->sens.value < 0)
        vifd_debug("SIOCGIWSENS : %i dB\n", vif_config->sens.value);
    else
        vifd_debug("SIOCGIWSENS : %i\n", vif_config->sens.value);

    return 0;
}

int
get_range(struct iwreq * request, struct wireless_info * vif_config)
{
    struct iw_range * range = (struct iw_range *) shmem_attach(iw_shm_id, NULL);
    memcpy(range, &(vif_config->range), sizeof(struct iw_range));
    request->u.data.length = sizeof(struct iw_range);
    vifd_debug("SIGCGIWRANGE : we_version_compiled = %i\n", range->we_version_compiled);

    return 0;
}

int
get_priv(struct iwreq * request, struct wireless_info * vif_config)
{
    int i;
    int num_calls = sizeof(vif_private_args) / sizeof(struct iw_priv_args);
    struct iw_priv_args * priv = \
        (struct iw_priv_args *) shmem_attach(iw_shm_id, NULL);

    vifd_debug("SIOCGIWPRIV : %i ioctl (%p):\n", num_calls, priv);
    for (i = 0; i < num_calls; i++)
    {
        memmove(&priv[i], &vif_private_args[i], sizeof(struct iw_priv_args));
    }
    request->u.data.length = num_calls;
    request->u.data.pointer = priv;
    for (i = 0; i < num_calls; i++)
    {
        vifd_debug("   (%p) cmd: %.4X\n", &(priv[i].cmd), priv[i].cmd);
        vifd_debug("   (%p) name: %s\n", &(priv[i].name), priv[i].name);
    }
    shmem_detach((char *) priv);

    return 0;
}

int
get_apaddr(struct iwreq * request, struct wireless_info * vif_config)
{
    char buffer[sizeof(struct sockaddr)];

    memmove(&(request->u.ap_addr.sa_data), &(vif_config->ap_addr.sa_data), \
        ETH_ALEN);
    vifd_debug("SIGCGIWAP : %s\n", iw_sawap_ntop(&(request->u.ap_addr), buffer));

    return 0;
}

int
set_essid(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)                     // essid settings (default: none)
    {
        vif_config->b.has_essid = 0;
        vif_config->b.essid_on = 1;
        vif_config->b.essid[0] = '\0';
    } else {
        char * essid = shmem_attach(iw_shm_id, NULL);
        strncpy(vif_config->b.essid, essid, request->u.essid.length);
        vif_config->b.essid_on = request->u.essid.flags;
        vif_config->b.has_essid = request->u.essid.flags;
        shmem_detach(essid);
    }
    vifd_debug("SIOCSIWESSID : \"%s\"\n", (char *) vif_config->b.essid);

    return 0;
}

int
get_essid(struct iwreq * request, struct wireless_info * vif_config)
{
    char * essid = shmem_attach(iw_shm_id, NULL);

    strncpy(essid, vif_config->b.essid, IW_ESSID_MAX_SIZE + 1);
    request->u.essid.pointer = essid;
    request->u.essid.length = strlen(essid) + 1;        // include trailing '\0'
    request->u.essid.flags = vif_config->b.essid_on;
    vifd_debug("SIOCGIWESSID : \"%s\", u.data.length = %i\n", \
        (char *) request->u.data.pointer, request->u.data.length);
    shmem_detach(essid);

    return 0;
}

int
get_nick(struct iwreq * request, struct wireless_info * vif_config)
{
    char * nickname = shmem_attach(iw_shm_id, NULL);

    // nickname uses same length as ESSID (see wireless.h)
    strncpy(nickname, vif_config->nickname, IW_ESSID_MAX_SIZE + 1);
    request->u.data.pointer = nickname;
    request->u.data.length = strlen(nickname) + 1;      // include trailing '\0'
    vifd_debug("SIOCGIWNICKN : \"%s\"\n", (char *) request->u.data.pointer);

    return 0;
}

int
set_bitrate(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)             // set initial bitrate (default: 0 Mbit/s)
    {
        vif_config->bitrate.value = 0;
        vif_config->bitrate.disabled = 1;
        vif_config->bitrate.fixed = 0;
    } else {
        memmove(&(vif_config->bitrate), &(request->u.bitrate), \
            sizeof(struct iw_param));
    }
    vif_config->has_bitrate = set_iw_param_state(&(vif_config->bitrate));
    vifd_debug("SIOCSIWRATE : %.1f Mbit/s\n", vif_config->bitrate.value / MEGA);

    return 0;
}

int
get_bitrate(struct iwreq * request, struct wireless_info * vif_config)
{
    memmove(&(request->u.bitrate), &(vif_config->bitrate), \
        sizeof(struct iw_param));
    vifd_debug("SIOCGIWRATE : %.1f Mbit/s\n", request->u.bitrate.value / MEGA);

    return 0;
}

int
set_rts(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)             // rts threshold in bytes (default: off)
    {
        vif_config->rts.value = 0;
        vif_config->rts.disabled = 1;
        vif_config->rts.fixed = 1;
    } else {
        memmove(&(vif_config->rts), &(request->u.rts), sizeof(struct iw_param));
    }
    vif_config->has_rts = set_iw_param_state(&(vif_config->rts));
    if (vif_config->has_rts == 1)
    {
            vifd_debug("SIOCSIWRTS : %i Bytes\n", request->u.rts.value);
    } else {
        vifd_debug("SIOCSIWRTS : off\n");
    }

    return 0;
}

int
get_rts(struct iwreq * request, struct wireless_info * vif_config)
{
    memmove(&(request->u.rts), &(vif_config->rts), sizeof(struct iw_param));
    if (vif_config->has_rts == 1)
    {
        vifd_debug("SIOCGIWRTS : %i Bytes\n", request->u.rts.value);
    } else {
        vifd_debug("SIOCGIWRTS : off\n");
    }

    return 0;
}

int
set_frag(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)  // set fragmentation threshold in bytes (default: off)
    {
        vif_config->frag.value = 0;
        vif_config->frag.disabled = 1;
        vif_config->frag.fixed = 1;
    } else {
        memmove(&(vif_config->frag), &(request->u.frag), sizeof(struct iw_param));
    }
    vif_config->has_frag = set_iw_param_state(&(vif_config->frag));
    if (vif_config->has_frag == 1)
    {
        vifd_debug("SIOCSIWFRAG : %i Bytes\n", request->u.frag.value);
    } else {
        vifd_debug("SIOCSIWFRAG : off\n");
    }

    return 0;
}

int
get_frag(struct iwreq * request, struct wireless_info * vif_config)
{
    memmove(&(request->u.frag), &(vif_config->frag), sizeof(struct iw_param));
    if (vif_config->has_frag == 1)
    {
        vifd_debug("SIOCGIWFRAG : %i Bytes\n", request->u.frag.value);
    } else {
        vifd_debug("SIOCGIWFRAG : off\n");
    }

    return 0;
}

int
set_txpower(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)
    {
        vif_config->txpower.value = 17;
        vif_config->txpower.disabled = 0;
        vif_config->txpower.flags = IW_TXPOW_DBM;
        vif_config->txpower.fixed = 1;
    } else {
        memmove(&(vif_config->txpower), &(request->u.txpower), \
            sizeof(struct iw_param));
    }
    vif_config->has_txpower = set_iw_param_state(&(vif_config->txpower));
    vifd_debug("SIOCSIWTXPOW : %d dBm\n", vif_config->txpower.value);

    return 0;
}

int
get_txpower(struct iwreq * request, struct wireless_info * vif_config)
{
    memmove(&(request->u.txpower), &(vif_config->txpower), \
        sizeof(struct iw_param));
    vifd_debug("SIOCGIWTXPOW : %d dBm\n", request->u.txpower.value);

    return 0;
}

int
set_retry(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)
    {
        vif_config->retry.value = 0;
        vif_config->retry.disabled = 1;
        vif_config->retry.fixed = 1;
    } else {
        memmove(&(vif_config->retry), &(request->u.retry), \
            sizeof(struct iw_param));
    }
    vif_config->has_retry = set_iw_param_state(&(vif_config->retry));
    vifd_debug("SIOCSIWRETRY : %.1d\n", request->u.retry.value);

    return 0;

}

int
get_retry(struct iwreq * request, struct wireless_info * vif_config)
{
    memmove(&(request->u.retry), &(vif_config->retry), sizeof(struct iw_param));
    vifd_debug("SIOCGIWRETRY : %.1d\n", request->u.retry.value);

    return 0;
}

int
set_encoding(struct iwreq * request, struct wireless_info * vif_config)
{
    if (request == NULL)    // initial key settings (default: encoding disabled)
    {
        vif_config->b.has_key = 0;
        vif_config->b.key_size = 0;
        vif_config->b.key[0] = '\0';
        vif_config->b.key_flags = IW_ENCODE_DISABLED;
    } else {
        unsigned char * key = (unsigned char *) shmem_attach(iw_shm_id, NULL);
        vif_config->b.key_size = request->u.encoding.length;
        memmove(vif_config->b.key, key, vif_config->b.key_size);
        vif_config->b.key_flags = request->u.encoding.flags;
        vif_config->b.has_key = 1;
        shmem_detach((char *) key);
    }
    vifd_debug("SIOCSIWENCODE : \"%s\"\n", vif_config->b.key);

    return 0;
}

int
get_encoding(struct iwreq * request, struct wireless_info * vif_config)
{
    if (vif_config->b.has_key == 1)
    {
        vifd_debug("SIOCGIWENCODE : <to implement!>\n");
        return IWERR_GET_EXT;
    }
    request->u.encoding.flags = IW_ENCODE_DISABLED;
    request->u.encoding.length = 0;
    request->u.encoding.pointer = NULL;
    vifd_debug("SIOCGIWENCODE : off\n");

    return 0;
}

int
get_power(struct iwreq * request, struct wireless_info * vif_config)
{
    memmove(&(request->u.power), &(vif_config->power), sizeof(struct iw_param));
    if (vif_config->has_power == 1)
    {
        vifd_debug("SIGCGIWPOWER : 'enabled'\n");
    } else {
        vifd_debug("SIGCGIWPOWER : 'disabled'\n");
    }

    return 0;
}
